//=============================================================================
// File Name: bipartite_solver.cc
// Written by: Yoram Singer & Shai Shalev-Shwartz
//=============================================================================

#ifndef _SHAI_BIPARTITE_CC_
#define _SHAI_BIPARTITE_CC_

// Included files
#include "bipartite_solver.h"

// some usefull macros
#define min_(x,y) ( ((x) < (y)) ? (x) : (y) )
#define max_(x,y) ( ((x) > (y)) ? (x) : (y) )

#define MISPAR_KATAN_KATAN 1e-15

/*----------------------------------------------------------------------------*/
void BipartiteSolver(infra::vector_view predictions,
		     infra::vector_view gamma,
		     std::vector<uint> &A,
		     std::vector<uint> &B,
		     double norm_x,double C,
		     infra::vector_view new_alpha) {

  // step 1: calculate Mu and Nu
  std::vector<double> Mu_; Mu_.clear();
  std::vector<double> Nu_; Nu_.clear();
  
  for (uint i=0; i<A.size(); ++i) {
    Mu_.push_back((gamma(A[i]) - predictions(A[i])) / norm_x);
  }
  for (uint i=0; i<B.size(); ++i) {
    Nu_.push_back((-gamma(B[i]) + predictions(B[i])) / norm_x);
  }
  
  // Sort the margins in a decreasing order...
  sort(Mu_.begin(), Mu_.end(), std::greater<double>());
  sort(Nu_.begin(), Nu_.end(), std::greater<double>());
  
  // Add pseudo-elements to avoid checking for termination conditions
  // during the main loop
  Mu_.push_back(Mu_[Mu_.size()-1]-C);
  Nu_.push_back(Nu_[Nu_.size()-1]-C);
  
  // On each iteration we maintain an semi-closed interval
  // of the form [c,next_c)
  double c = 0.0;
  
  // The proposed number of "active" margins. r & s start at 1 and can be as
  // large as the number of margin values (size of Mu and Nu respectively)
  int r = 1, s = 1;
  
  // The cumulative sum of the Mu margins up to the index r
  double sum_mu = Mu_[0];
  
  // The cumulative sum of the Nu margins up to the index s
  double sum_nu = Nu_[0];
  
  // The default best value for c (the actual sum of alpha and beta) is set
  // to the upper bound C 
  double best_c = C;

  while (c < C) {
    // Compute the optimal value of c under the assumption that the number
    // of active margins is r & s
    double copt = (s * sum_mu + r * sum_nu) / ((double) (r + s));
    
    // Compute the potential next grid point for c by increasing r
    double next_cr = (sum_mu + Mu_[r]) - (r + 1) * Mu_[r];
    
    // Compute the potential next grid point for c by increasing s
    double next_cs = (sum_nu + Nu_[s]) - (s + 1) * Nu_[s];

    // The right point of the interval cannot exceed C
    double next_c = min_(min_(next_cr, next_cs), C);
    
    // If the optimal value indeed falls in [c,next_c) we are done
    if ( c <= copt && copt < next_c ) {
      best_c = copt;
      break;
    }
    
    // If we hit C we must stop and set the optimum to be C
    if ( next_c >= C - MISPAR_KATAN_KATAN ) {
      best_c = C;
      break;
    }
    
    // Update the candidates for next_c 
    if ( next_cr < next_cs ) {
      sum_mu += Mu_[r];
      r++;
    } else {
      sum_nu += Nu_[s];
      s++;
    }
    
    // Switch from the interval [c,next_c) to the interval [next_c,?)
    c = next_c;
  }
  
  // Calculate theta for both alpha and beta
  double theta_a = (sum_mu - best_c) / ((double) r);
  double theta_b = (sum_nu - best_c) / ((double) s);
  
  // Update alpha : for i in A we need to add alpha and for i in B
  // we need to substract beta and we define alpha to be -beta. 
  for (uint i=0; i < A.size(); ++i) {
    new_alpha(A[i]) = max_(0.0, (gamma(A[i]) - predictions(A[i])) / norm_x - theta_a );
  }
  for (uint i=0; i < B.size(); ++i) {
    new_alpha(B[i]) = - max_(0.0, (-gamma(B[i]) + predictions(B[i])) / norm_x - theta_b);
  }

//   std::cerr << "theta_a = " << theta_a << " theta_b = " << theta_b 
// 	    << " C = " << C << " A[0] = " << A[0] << ", new_alpha = " << new_alpha << std::endl;

  
}



#endif
