//=============================================================================
// File Name: mcDense.cc
// Written by: Shai Shalev-Shwartz (23.03.06)
// This code implements three multiclass online learning algorithm:
//   - Perceptron with max-update
//   - PA-I with max-update
//   - SOPOPO for multiclass
//=============================================================================
#ifndef _SHAI_MCDENSE_
#define _SHAI_MCDENSE_

#define EXPECTED_ERROR_RATIO 0.10
#define INITIAL_SIZE_LIMIT   10000

//*****************************************************************************
// Included Files
//*****************************************************************************
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstdlib>
#include <cmdline/cmd_line.h>
#include <kernels/kernels.h>
#include <kernels/kernels.imp>
#include <infra.h>
#include "bipartite_solver.h"

#define _min(x,y) ( ((x) < (y)) ? (x) : (y) )

int main(int argc, char** argv) {

  // PARSE COMMAND LINE
  // ====================================================================
  // check for the correct number of command line arguments and print usage
  // message
  std::string data_filename;
  std::string label_filename;
  std::string kernel_filename;
  std::string method;
  double C;
  std::string out_file;
  std::string log_file_name;
  uint max_examples;
  uint num_labels;

  // parse command line
  learning::cmd_line cmdline;
  cmdline.info("Online multiclass learning");
  cmdline.add_master_option("<instance file name (infra format)>", 
			    &data_filename);
  cmdline.add_master_option("<labels file name (ascii numbers from 0,...,k-1)>", 
			    &label_filename);
  cmdline.add("-C", "value of the C parameter [default 1.0]", &C, 1.0);
  cmdline.add("-Kernel", "Name of text file with the kernel name and then an infra "\
	      "vector of parameters (infra txt format)  \n\t\t[if this is not "\
	      "specified, an RBF kernel with sigma=1.0 is used]", 
	      &kernel_filename, "default");
  cmdline.add("-m","maximal number of examples to read from the file\n "\
	      "\t[default ten milion (10000000)]",
	      &max_examples,
	      10000000);
  cmdline.add("-log","Name of a file in which on each line there is the number of trial,\n "\
	      "\tthe prediction, the number of cumulative mistakes and cumulative loss.\n"\
	      "\t[default no_log]",
	      &log_file_name,
	      "no_log");
  cmdline.add("-method", "The online method to use: \n\tPerc for Perceptron, "\
	      "PA for Passive-Aggressive, "\
	      "Sop for SOPOPO\n [Default: PA]",
	      &method,
	      "PA");

  int rc = cmdline.parse(argc, argv);
  if (rc < 2) {
    cmdline.print_help();
    return EXIT_FAILURE;
  }
  if (method != "Perc" && method != "PA" && method != "Sop") {
    std::cerr << "incorrect method type. Abort " << std::endl;
    return EXIT_FAILURE;
  }
  


  // CONSTRUCT A KERNEL
  // ==================
  kernels::KernelsBuilder K(kernel_filename);


  // OPEN DATA AND LABEL FILES
  // =========================
  FILE* data_file = fopen(data_filename.c_str(),"r");
  std::ifstream label_file(label_filename.c_str());
  if (!label_file.good()) {
    std::cerr << "error w/ label_file" << std::endl;
    return EXIT_FAILURE;
  }


  // open log file
  FILE * log_fid = NULL;
  if (log_file_name != "no_log")
    log_fid = fopen(log_file_name.c_str(),"wt");


  // INITIALIZATIONS
  // ===============

  // read the number of lables. This is the first number in the labels file
  label_file >> num_labels;

  // read the number of instances(m) and their dimension(n). The first two 
  // entries in the data file are m and n.
  unsigned long m,n;
  infra::load_binary(data_file, m);
  infra::load_binary(data_file, n);

  if (m > max_examples) m = max_examples; // don't read more than max_examples

  // allocate a matrix for holding the support and allocate a matrix for 
  // holding the alphas. 
  unsigned long init_size = (unsigned long)(m / 10);
  if ((init_size) < 100) init_size = 100;
  infra::matrix S(init_size,n); // support matrix
  infra::matrix alpha(init_size,num_labels); // weights
  unsigned long svc = 0; // the number of support vectors so far

  // Count mistakes, and loss
  unsigned long mistakes = 0;
  double loss = 0.0;


  // BEGIN ONLINE LEARNING
  // =====================
  for(unsigned long rounds=0; rounds<m; ++rounds) {

    // Receive a new instance from the data file
    // =========================================
    infra::vector instance(data_file);   

    // predict a label 
    // ===============
    infra::matrix_view cur_support = S.submatrix(0,0,svc,n);
    infra::matrix_view cur_alpha = alpha.submatrix(0,0,svc,num_labels);
    infra::vector_view S_mult_x = K(cur_support,instance);
    infra::vector_view ranks =  S_mult_x * cur_alpha;
    uint predicted_label = ranks.argmax();

    // Read the correct label from the labels file
    // ===========================================
    uint correct_label;
    label_file >> correct_label;

    // calculate the loss
    // ===================

    // first, find the highest incorrect label
    uint highest_ranked_incorrect = predicted_label;
    if (correct_label == predicted_label) {
      double tmp = ranks(correct_label);
      ranks(correct_label) = ranks.min() - 1000000.0;
      highest_ranked_incorrect = ranks.argmax();
      ranks(correct_label) = tmp;
    }
    else {
      mistakes++;
    }
    
    // calculate the margin of the classification
    double margin = ranks(correct_label) - ranks(highest_ranked_incorrect);

    // and the loss is
    double cur_loss = 1 - margin; 
    if (cur_loss < 0.0) cur_loss = 0.0;

    loss += cur_loss; // the cumulative loss

    // update
    // ======
    if((method != "Perc" && cur_loss > 0.0) || (cur_loss >= 1.0)) {
      
      // make roon if necessary for the new support vector
      if (svc == S.height()) { 
	unsigned long new_size = svc*2;
	if (new_size > m) new_size = m;
	alpha.resize(new_size,num_labels);
	S.resize(new_size,n);
      }

      // add the new support 
      S.row(svc) = instance;
      
      // calculate the new alpha
      infra::vector_view cur_alpha = alpha.row(svc);
      if (method == "Perc" || method == "PA") {

	double weight = 1.0; // Perceptron
	if (method == "PA") { // PA-I
	  weight = cur_loss / (2.0*K(instance,instance));
	  if (weight > C) weight = C;
	} 

	// Perceptron and PA update
	cur_alpha.zeros();
	cur_alpha(correct_label) = weight;
	cur_alpha(highest_ranked_incorrect) = -weight;

      } else {

	// Sopopo update
	infra::vector gamma(num_labels); gamma.zeros();
	gamma(correct_label) = 1.0;
	std::vector<uint> A; A.clear(); A.push_back(correct_label);
	std::vector<uint> B; B.clear(); 
	for (uint j=0; j < num_labels; ++j) {
	  if (j != correct_label) 
	    B.push_back(j);
	}
	BipartiteSolver(ranks,gamma,A,B,K(instance,instance),
			C,cur_alpha);

      }

      // increment the support size variable
      ++svc;

    } 

    // print some statistics
    if (log_fid != NULL)
      fprintf(log_fid,"rounds: %lu mistakes: %lu loss %.3f svc %lu\n",
	      rounds,mistakes,loss,svc);


  }

  // PRINT OUT THE FINAL REPORT (do not change without discussing)
  // ============================================================
  std::cout << "method: " << method
	    << "\tC: " << C
	    << "\trounds: " << m
	    << "\tmistakes: " << mistakes
	    << "\tloss: " << loss
	    << "\terror_rate: " << double(mistakes)/double(m)
    	    << "\tavg_loss: " << loss/double(m) 
            << "\trep_set_size: " << svc
	    << std::endl; 


  // close log file
  if (log_fid != NULL) 
    fclose(log_fid);

  return(EXIT_SUCCESS);

}


#endif
//*****************************************************************************
//                                     E O F
//*****************************************************************************

